/////////////////////////////////////////////////////////////////////////////
//
// Written by Igor Tolmachev (support@itsamples.com)
// Copyright (c) 2017 Igor Tolmachev.
//
// This code may be used in compiled form in any way you desire. This
// source file may be redistributed by any means PROVIDING it is 
// not sold for profit without the authors written consent, and 
// providing that this notice and the authors name is included. 
//
// This file is provided "as is" with no expressed or implied warranty.
// The author accepts no liability if it causes any damage to you or your
// computer whatsoever. It's free, so don't hassle me about it.
//
// Beware of bugs.
//

#include "stdafx.h"
#include "FileTreeCtrl.h"
#include "AtlFile.h"

CSystemImageList::CSystemImageList()
{
    if (!m_ImageList)
    {
        m_ImageList.Create(16, 16, ILC_COLOR32, 0, 1);
        m_ImageList.SetBkColor(RGB(255, 255, 255));

        AddSystemIcon(CSIDL_DESKTOP);
        AddSystemIcon(CSIDL_APPDATA);
        AddSystemIcon(CSIDL_BITBUCKET);

        SHFILEINFO sfi;
        ZeroMemory(&sfi, sizeof(SHFILEINFO));
        SHGetFileInfo((LPCTSTR)_T("c:\\"),  0, &sfi, sizeof(SHFILEINFO), SHGFI_SYSICONINDEX | SHGFI_SMALLICON);

        HICON hIconDefault = LoadIcon(_Module.m_hInst, MAKEINTRESOURCE(IDI_ICON_FILE));

        AddFileTypeIcon(&sfi, _T(".mui"), hIconDefault);
        AddFileTypeIcon(&sfi, _T(".dat"), hIconDefault);
        AddFileTypeIcon(&sfi, _T(".log"), hIconDefault);
        AddFileTypeIcon(&sfi, _T("explorer.exe"), hIconDefault);
        AddFileTypeIcon(&sfi, _T(".dll"), hIconDefault);
        AddFileTypeIcon(&sfi, _T(".ttf"), hIconDefault);
        AddFileTypeIcon(&sfi, _T(".ini"), hIconDefault);
    }
}

void CSystemImageList::AddSystemIcon(DWORD dwFolder)
{
    SHFILEINFO sfi;
    ZeroMemory(&sfi, sizeof(SHFILEINFO));

    LPITEMIDLIST pidl;
    if (SUCCEEDED(SHGetFolderLocation(0, dwFolder, 0, 0, &pidl)))
    {
#ifdef _UNICODE
        if (SHGetFileInfo((LPCWSTR)pidl, 0, &sfi, sizeof(sfi), SHGFI_ICON | SHGFI_PIDL))
#else
        if (SHGetFileInfo((LPCSTR)pidl,  0, &sfi, sizeof(sfi), SHGFI_ICON | SHGFI_PIDL))
#endif
            m_ImageList.AddIcon(sfi.hIcon);

        CoTaskMemFree(pidl);
    }
    else
        m_ImageList.AddIcon(LoadIcon(_Module.m_hInst, MAKEINTRESOURCE(IDI_ICON_FOLDER)));
}

void CSystemImageList::AddFileTypeIcon(SHFILEINFO * pSfi, LPCTSTR lpExt, HICON hDefault)
{
    if (SHGetFileInfo(lpExt, FILE_ATTRIBUTE_NORMAL, pSfi, sizeof(SHFILEINFO), SHGFI_ICON | SHGFI_SMALLICON | SHGFI_USEFILEATTRIBUTES) != 0)
        m_ImageList.AddIcon(pSfi->hIcon);
    else
        m_ImageList.AddIcon(hDefault);
}

CSystemImageList::~CSystemImageList()
{
    if(m_ImageList)
    {
        int nSize = m_ImageList.GetImageCount();
        for(int i = 0; i < nSize; i++)
        {
            HICON hIcon = m_ImageList.GetIcon(i);
            if(hIcon)
                DestroyIcon(hIcon);
        }

        m_ImageList.RemoveAll();
    }
}

CImageList* CSystemImageList::GetImageList()
{
    return &m_ImageList;
}

BOOL CFileTreeCtrl::PreTranslateMessage(MSG* pMsg)
{
    return FALSE;
}

BOOL CFileTreeCtrl::SubclassWindow(HWND hWnd)
{
    BOOL bResult = CWindowImpl<CFileTreeCtrl, CTreeViewCtrl>::SubclassWindow(hWnd);
    if(bResult)
        PostMessage(WM_TREE_BUILD);

    return bResult;
}

BOOL CFileTreeCtrl::IsEqual(CString strValue1, CString strValue2)
{
    strValue1.MakeLower();
    strValue2.MakeLower();

    return (strValue1 == strValue2);
}

CString CFileTreeCtrl::GetSelectedPath()
{
    return GetItemPath(GetSelectedItem());
}

void CFileTreeCtrl::OnTreeRefresh()
{
    SetRedraw(FALSE);

    HTREEITEM hSelItem = GetSelectedItem();
    CString strItem  = GetItemPath(hSelItem);

    if(strItem.IsEmpty())
        DisplayRoot();
    else
        DisplayPath(strItem, hSelItem);

    SetRedraw(TRUE);
}

CString CFileTreeCtrl::GetItemPath(HTREEITEM hItem)
{
    CString strPath;

    HTREEITEM hParent = hItem;
    CString strItem = "";
    int nLength = 0;
    TCHAR szItemText[_MAX_PATH] = { 0 };

    while(hParent)
    {
        GetItemText(hParent, szItemText, _MAX_PATH);
        strItem = szItemText;
        nLength = strItem.GetLength();
        hParent = GetParentItem(hParent);

        if(strItem[nLength - 1] == _T('\\'))
            strPath = strItem + strPath;
        else
        {
            if(strPath.GetLength() > 0)
                strPath = strItem + _T('\\') + strPath;
            else
                strPath = strItem;
        }
    }

    if(m_strRootFolder.GetLength() > 0)
        strPath = m_strRootFolder + _T('\\') + strPath;

    return strPath;
}

BOOL CFileTreeCtrl::IsFolder(HTREEITEM hItem)
{
    TVITEM item = { 0 };
    item.mask = TVIF_PARAM;
    item.hItem = hItem;
    if(GetItem(&item))
        return (item.lParam > 0);

    return FALSE;
}

int CFileTreeCtrl::GetIconIndex(HTREEITEM hItem)
{
    TV_ITEM tvi;
    ZeroMemory(&tvi, sizeof(TV_ITEM));
    tvi.mask   = TVIF_IMAGE;
    tvi.hItem  = hItem;
    if(GetItem(&tvi))
        return tvi.iImage;
    else
        return -1;
}

int CFileTreeCtrl::GetIconIndex(const CString strFileName, bool bIsFolder)
{
    CString strLwrFileName(strFileName);
    strLwrFileName.MakeLower();

    if(bIsFolder)
    {
        if (strLwrFileName.Find(_T("recycle")) != -1)
            return 2;
        else
            return 1;
    }
    else
    {
        if (strLwrFileName.Find(_T(".mui")) != -1)
            return 3;
        else if (strLwrFileName.Find(_T(".dat")) != -1)
            return 4;
        else if (strLwrFileName.Find(_T(".log")) != -1)
            return 5;
        else if (strLwrFileName.Find(_T(".exe")) != -1 || strLwrFileName.Find(_T("bootmgr")) != -1)
            return 6;
        else if (strLwrFileName.Find(_T(".dll")) != -1)
            return 7;
        else if (strLwrFileName.Find(_T(".ttf")) != -1)
            return 8;
        else
            return 9;
    }
}

BOOL CFileTreeCtrl::HasChildEntries(const CString strDirectory)
{
    CString strFile;
    if(strDirectory[strDirectory.GetLength() - 1] == _T('\\'))
        strFile = strDirectory + _T("*.*");
    else
        strFile = strDirectory + _T("\\*.*");

    CFindFile find;
    BOOL bFind = find.FindFile(strFile);
    while(bFind)
    {
        if((find.IsDirectory() && !find.IsDots()) || !find.IsDirectory())
            return TRUE;

        bFind = find.FindNextFile();
    }

    return FALSE;
}

HTREEITEM CFileTreeCtrl::InsertFileItem(const CString strFile, const CString strPath, bool bIsFolder, HTREEITEM hParent)
{
    CString strFullPath = strPath;
    if(strFullPath[strFullPath.GetLength() - 1] != '\\')
        strFullPath += _T("\\");
    strFullPath += strFile;

    int nIconIndex = 0;
    if(hParent != TVI_ROOT)
        nIconIndex = GetIconIndex(strFullPath, bIsFolder);

    CString strText;
    if(!strFile.IsEmpty())
        strText = strFile;
    else
        strText = strPath;

    TV_INSERTSTRUCT tvis;
    ZeroMemory(&tvis, sizeof(TV_INSERTSTRUCT));
    tvis.hParent             = hParent;
    tvis.hInsertAfter        = TVI_LAST;
    tvis.item.mask           = TVIF_CHILDREN | TVIF_IMAGE | TVIF_PARAM | TVIF_SELECTEDIMAGE | TVIF_TEXT;
    tvis.item.lParam         = bIsFolder;
    tvis.item.iImage         = nIconIndex;
    tvis.item.cChildren      = HasChildEntries(strFullPath);
    tvis.item.iSelectedImage = nIconIndex;

#ifdef _UNICODE
    tvis.item.pszText        = (LPWSTR)strText.GetBuffer(0);
#else
    tvis.item.pszText        = (LPSTR)strText.GetBuffer(0);
#endif
    strText.ReleaseBuffer();

    return InsertItem(&tvis);
}

void CFileTreeCtrl::DisplayRoot()
{
    CWaitCursor waiter;

    DeleteAllItems();

    TCHAR szVName[MAX_PATH + 1];
    TCHAR szPaths[1024];
    TCHAR szFName[MAX_PATH + 1];
    BOOL bFirstTime = TRUE;
    BOOL bFindDrive = FALSE;
    DWORD dwLen = 0;
    ULARGE_INTEGER ulTotalBytes;

    SetRedraw(FALSE);

    // Get the first volume (handle is used to iterate)
    HANDLE hVolume = FindFirstVolume(szVName, ARRAYSIZE(szVName));
    if (hVolume != INVALID_HANDLE_VALUE)
    {
        // Iterate through all volumes
        while (bFirstTime || FindNextVolume(hVolume, szVName, ARRAYSIZE(szVName)))
        {
            bFirstTime = FALSE;

            // Make sure the volume path is valid
            dwLen = (DWORD)_tcslen(szVName);
            if (dwLen < 5 || _tcsncmp(szVName, _T("\\\\?\\"), 4) != 0 || szVName[dwLen - 1] != L'\\')
                continue;

            // Check the drive type (must be FIXED)
            if (GetDriveType(szVName) != DRIVE_FIXED)
                continue;

            // Make sure there are no paths (it is hidden)
            if (!GetVolumePathNamesForVolumeName(szVName, szPaths, ARRAYSIZE(szPaths), &dwLen) || szPaths[0] != 0)
                continue;

            // Make sure the label is System Reserved and filesystem is NTFS
            if (!GetVolumeInformation(szVName, NULL, 0, NULL, NULL, NULL, szFName, ARRAYSIZE(szFName)) ||
                _tcscmp(szFName, _T("NTFS")) != 0) 
                continue;

            // Accept any size less than 500MB
            ulTotalBytes.QuadPart = 0;
            if (!GetDiskFreeSpaceEx(szVName, NULL, &ulTotalBytes, NULL) || ulTotalBytes.QuadPart > 524288000LL)
                continue;

            FindVolumeClose(hVolume);

            InsertFileItem(_T(""), szVName, true, TVI_ROOT);
            bFindDrive = TRUE;
            break;
        }

        if(!bFindDrive)
            MessageBox(_T("The hidden system drive was not found"), _T("SRV Explorer"), MB_ICONWARNING | MB_OK);
    }
    else
        MessageBox(_T("The hidden system drive was not found"), _T("SRV Explorer"), MB_ICONWARNING | MB_OK);

    SetRedraw(TRUE);
}

void CFileTreeCtrl::DisplayPath(const CString strPath, HTREEITEM hParent)
{
    CWaitCursor waiter;

    SetRedraw(FALSE);

    // Remove all the items currently under hParent
    HTREEITEM hChild = GetChildItem(hParent);
    while(hChild)
    {
        DeleteItem(hChild);
        hChild = GetChildItem(hParent);
    }

    // Find all the directories and files underneath strPath
    CSimpleArray <CString> vDirectoryPaths;
    CSimpleArray <CString> vFilePaths;

    CFindFile find;
    CString strFile;
    CString strNewPath;
    if(strPath[strPath.GetLength() - 1] == _T('\\'))
        strFile = strPath + _T("*.*");
    else
        strFile = strPath + _T("\\*.*");

    DWORD dwError = NULL;
    BOOL bFind = find.FindFile(strFile.GetBuffer(0));
    while(bFind)
    {
        if(find.IsDirectory())
        {
            if(!find.IsDots())
            {
                strNewPath = find.GetFilePath();
                vDirectoryPaths.Add(strNewPath);
            }
        }
        else
        {
            strNewPath = find.GetFilePath();
            vFilePaths.Add(strNewPath);
        }

        bFind = find.FindNextFile();
    }

    TCHAR szPathBuffer[_MAX_PATH];
    TCHAR szFileName[_MAX_FNAME];
    TCHAR szExtension[_MAX_EXT];

    int i = 0;
    for(i = 0; i < vDirectoryPaths.GetSize(); ++i)
    {
        _tsplitpath_s(vDirectoryPaths[i].GetBuffer(0), NULL, NULL, NULL, NULL, szFileName, _MAX_FNAME, szExtension, _MAX_EXT);
        vDirectoryPaths[i].ReleaseBuffer();
        _tmakepath_s(szPathBuffer, NULL, NULL, szFileName, szExtension);
        InsertFileItem(szPathBuffer, strPath, true, hParent);
    }

    for(i = 0; i < vFilePaths.GetSize(); ++i)
    {
        _tsplitpath_s(vFilePaths[i].GetBuffer(0), NULL, NULL, NULL, NULL, szFileName, _MAX_FNAME, szExtension, _MAX_EXT);
        vFilePaths[i].ReleaseBuffer();
        _tmakepath_s(szPathBuffer, NULL, NULL, szFileName, szExtension);
        InsertFileItem(szPathBuffer, strPath, false, hParent);
    }

    SetRedraw(TRUE);
}

LRESULT CFileTreeCtrl::OnTreeBuilding(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    SetImageList(*(m_SysImageList.GetImageList()), TVSIL_NORMAL);

    OnTreeRefresh();

    return 0;
}

LRESULT CFileTreeCtrl::OnLButtonDblClick(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    if(!IsFolder(GetSelectedItem()))
    {
        CString strPath = GetSelectedPath();

        ::SendMessage(GetParent(), WM_TREE_SELECT, 0, (LPARAM)strPath.GetBuffer(0));
        strPath.ReleaseBuffer();
    }

    return 0;
}

LRESULT CFileTreeCtrl::OnRButtonUp(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    ::SendMessage(m_hWnd, WM_LBUTTONDOWN, wParam, lParam);

    CString strPath = GetSelectedPath();

    UINT nMenuID = IDR_MENU_FILE;
    if(IsFolder(GetSelectedItem()))
        nMenuID = IDR_MENU_FOLDER;

    CMenu menu;
    if(menu.LoadMenu(nMenuID))
    {
        CMenuHandle menuPopup = menu.GetSubMenu(0);
        if(menuPopup.m_hMenu)
        {
            CPoint pt(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
            ClientToScreen(&pt);

            CString strPath = GetSelectedPath();

            int nCommand = menuPopup.TrackPopupMenu(TPM_RETURNCMD|TPM_LEFTALIGN|TPM_TOPALIGN|TPM_LEFTBUTTON, pt.x, pt.y, m_hWnd);
            switch(nCommand)
            {
            case ID_MENU_VIEW:
                ::SendMessage(GetParent(), WM_TREE_SELECT, 0, (LPARAM)strPath.GetBuffer(0));
                strPath.ReleaseBuffer();
                break;
            case ID_MENU_COPY:
                if(OpenClipboard())
                {
                    size_t nLen = strPath.GetLength();
                    HGLOBAL hMem =  GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, (nLen + 1) * sizeof(TCHAR));
                    memcpy(GlobalLock(hMem), strPath.GetBuffer(0), nLen * sizeof(TCHAR));
                    strPath.ReleaseBuffer();
                    GlobalUnlock(hMem);

                    if(EmptyClipboard())
#ifdef _UNICODE
                        SetClipboardData(CF_UNICODETEXT, hMem);
#else
                        SetClipboardData(CF_TEXT, hMem);
#endif

                    CloseClipboard();
                }
                break;
            case ID_MENU_SAVE:
                {
                    TCHAR szItemText[_MAX_PATH] = { 0 };
                    GetItemText(GetSelectedItem(), szItemText, _MAX_PATH);

                    CFileDialog fileDlg(false, NULL, szItemText, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT, _T("All Files\0*.*\0"));
                    if(IDOK == fileDlg.DoModal())
                    {
                        CFile fileSource;
                        if(fileSource.Open(strPath))
                        {
                            DWORD dwLen = fileSource.GetSize();
                            if(dwLen > 0)
                            {
                                LPBYTE lpBuffer = new BYTE[dwLen + 1];
                                memset(lpBuffer, 0, dwLen);

                                DWORD dwReadLen = 0;
                                if(fileSource.Read(lpBuffer, dwLen, &dwReadLen))
                                {
                                    CFile fileTarget;
                                    if(fileTarget.Create(fileDlg.m_szFileName))
                                    {
                                        fileTarget.Write(lpBuffer, dwReadLen);
                                        fileTarget.Close();
                                    }
                                    else
                                        MessageBox(_T("Can't open target file"), _T("SRV Explorer"), MB_ICONWARNING | MB_OK);
                                }

                                delete [] lpBuffer;
                            }
                            else
                                MessageBox(_T("Selected file is empty"), _T("SRV Explorer"), MB_ICONWARNING | MB_OK);

                            fileSource.Close();
                        }
                        else
                            MessageBox(_T("Can't open selected file"), _T("SRV Explorer"), MB_ICONWARNING | MB_OK);
                    }
                }
                break;
            case ID_MENU_PROPS:
                {
                    SHELLEXECUTEINFO info = { 0 };
                    info.cbSize = sizeof info;
#ifdef _UNICODE
                    info.lpFile = (LPCWSTR)strPath;
#else
                    info.lpFile = (LPCSTR)strPath;
#endif
                    info.nShow = SW_SHOW;
                    info.fMask = SEE_MASK_INVOKEIDLIST;
                    info.lpVerb = _T("properties");

                    ShellExecuteEx(&info);
                }
                break;
            }
        }
    }

    ::SendMessage(m_hWnd, WM_LBUTTONUP, wParam, lParam);

    return 0;
}

LRESULT CFileTreeCtrl::OnItemExpanding(int idCtrl, LPNMHDR pnmh, BOOL& bHandled)
{
    NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pnmh;
    if(pNMTreeView->action == TVE_EXPAND)
    {
        if(!GetChildItem( pNMTreeView->itemNew.hItem))
            DisplayPath(GetItemPath(pNMTreeView->itemNew.hItem), pNMTreeView->itemNew.hItem);
    }

    return 0;
}
